library(quantreg)

robbins_monro_interval <- function(estimate, range_l, range_u, start, alpha, steps, sampling_func)
{
  step.size.constant <- 1
  current.endpoint <- rnorm(1, mean=start)

  z <- qnorm(1-alpha)
  k <- (sqrt(2*pi)*2)/(z*exp((-z^2)/2))

  for(i in 1:steps)
  {
    current.endpoint <- min(range_u, current.endpoint)
    current.endpoint <- max(range_l, current.endpoint)

    current.estimate <- sampling_func(current.endpoint)

    y <- current.estimate < estimate

    current.endpoint <- current.endpoint + step.size.constant*(y-alpha)/i

    step.size.constant <- k*abs(current.endpoint - estimate)
  }

  return(current.endpoint)
}


regular_regression <- function(estimate, range_l, range_u, start, alpha, steps, sampling_func)
{
 # middle <- rnorm(1, mean=start)
  middle <- start
  points <- seq(range_l,range_u, length.out=steps)
  bootstrap.samples <- sapply(points, FUN=sampling_func)

  fit <- rq(bootstrap.samples ~ points, tau=alpha)

  intercept = fit$coefficients[1]
  slope = fit$coefficients[2]

  return((estimate-intercept)/slope)
}


## runs adaptive quantile regression and returns the estimated bound of the confidence interval, U.
## estimate is the original estimate of the parameter of interest.
## range_l is the minimal possible (or likely) value of the parameter
## range_u is the maximal possible (or likely) value of the parameter. 
## start is a guess of the value of U. Currently ignored, but needed for compitability with other functions.
## alpha is the confidence level.
## steps is the number of iterations to run for.
## sampling_func is an R function that recived a value for the parameter and returns a bootstrap estiamte of it.
adaptive_regression <- function(estimate, range_l, range_u, start, alpha, steps, sampling_func)
{
  starting.points <- 30
  #middle <- rnorm(1, mean=start)
  middle <- start
  points <- seq(range_l, range_u, length.out=starting.points)

  bootstrap.samples <-  sapply(points, FUN=sampling_func)

  fit <- rq(bootstrap.samples ~ points, tau=alpha)

  intercept = fit$coefficients[1]
  slope = fit$coefficients[2]

  current_endpoint <- (estimate-intercept)/slope

  for(i in 1:(steps-starting.points))
  {
    new.point <- (starting.points+i)*current_endpoint - sum(points)
    new.point <- min(range_u, new.point)
    new.point <- max(range_l, new.point)

    #cat("new point: ", new.point, '\n')

    new.bootstrap.sample <- sampling_func(new.point)

    points <- c(points, new.point)
    bootstrap.samples <- c(bootstrap.samples, new.bootstrap.sample)


    fit <- rq(bootstrap.samples ~ points, tau=alpha)

    intercept = fit$coefficients[1]
    slope = fit$coefficients[2]

    current_endpoint <- (estimate-intercept)/slope
    #cat("current endpoint is: ", current_endpoint, '\n')
  }

  return(current_endpoint)
}

efficient_robbins_monro <- function(estimate, range_l, range_u, start, alpha, steps, sampling_func)
{
  c <- 1
  alpha <- 1-alpha

  beta <- c/dnorm(qnorm(alpha))

  tau = 1
  current.endpoint <- rnorm(1, start)

  v = beta*tau

  z <- qnorm(alpha)
  k <- (sqrt(2*pi)*2)/(z*exp((-z^2)/2))

  #print(k)

  for(i in 1:steps)
  {
    #print(current.endpoint)
    current.endpoint <- min(range_u, current.endpoint)
    current.endpoint <- max(range_l, current.endpoint)

    current.estimate <- sampling_func(current.endpoint)

    y <- current.estimate > estimate

    b <- pnorm(qnorm(alpha)/sqrt(1+v))
    c <- (v/sqrt(1+v))*dnorm(qnorm(alpha)/sqrt(1+v))

    current.endpoint <- current.endpoint - c*(y-b)/(beta*b*(1-b))

    v <- v - c^2/(b*(1-b))

    c <- k*abs(current.endpoint - estimate)

    beta <- 1/(c*(dnorm(qnorm(alpha))))

  }

  return(current.endpoint)
}

simulation <- function(reps=500, FUN)
{
  results <- c()
  for(i in 1:reps)
  {
    result <- FUN()
    results <- c(results, result)
  }

  return(results)
}

simulation <- function(FUN, estimate, range_l, range_u, start, alpha, steps, sampling_func, reps)
{
  results <- c()
  for(i in 1:reps)
  {
    result <- FUN(estimate, range_l, range_u, start, alpha, steps, sampling_func)
    results <- c(results, result)
  }

  return(results)
}

calculate_mse_from_simulation <- function(method_func, estimate, range_l, range_u, start, correct_endpoint, alpha,steps, sampling_func,reps)
{
  results <- simulation(method_func, estimate, range_l, range_u, start, alpha, steps, sampling_func, reps)

  simulation.results <- data.frame(method=NULL, error=NULL, error_type=NULL)
  return(sqrt(sum((results-correct_endpoint)^2)/length(results)))
}


rm_regression <- function(estimate, range_l, range_u, start, alpha, steps, sampling_func)
{
  points <- seq(range_l,range_u, length.out=20)
  bootstrap.samples <- sapply(points, FUN=sampling_func)

  fit <- rq(bootstrap.samples ~ points + I(points^2), tau=alpha)
  c <- fit$coefficients

  roots <- suppressWarnings(as.numeric(polyroot(c)))

  root <- roots[(roots>range_l)&&(roots < range_u)]

  current.endpoint <- root

  z <- qnorm(1-alpha)
  k <- (sqrt(2*pi)*2)/(z*exp((-z^2)/2))

  step.size.constant <- 1

  for(i in 20:steps)
  {
    current.endpoint <- min(range_u, current.endpoint)
    current.endpoint <- max(range_l, current.endpoint)

    current.estimate <- sampling_func(current.endpoint)

    y <- current.estimate < estimate

    current.endpoint <- current.endpoint + step.size.constant*(y-alpha)/i

    step.size.constant <- k*abs(current.endpoint - estimate)
  }

  return(current.endpoint)
}

efficient_rm_regression <- function(estimate, range_l, range_u, start, alpha, steps, sampling_func)
{
  points <- seq(range_l,range_u, length.out=20)
  bootstrap.samples <- sapply(points, FUN=sampling_func)

  fit <- rq(bootstrap.samples ~ points + I(points^2), tau=alpha)
  c <- fit$coefficients

  roots <- suppressWarnings(as.numeric(polyroot(c)))

  root <- roots[(roots>range_l)&&(roots < range_u)]

  current.endpoint <- root

  z <- qnorm(1-alpha)
  k <- (sqrt(2*pi)*2)/(z*exp((-z^2)/2))

  c <- 1
  alpha <- 1-alpha

  beta <- c/dnorm(qnorm(alpha))

  tau = 0.1

  v = beta*tau

  for(i in 1:(steps-20))
  {
    #print(current.endpoint)
    current.endpoint <- min(range_u, current.endpoint)
    current.endpoint <- max(range_l, current.endpoint)

    current.estimate <- sampling_func(current.endpoint)

    y <- current.estimate > estimate

    b <- pnorm(qnorm(alpha)/sqrt(1+v))
    c <- (v/sqrt(1+v))*dnorm(qnorm(alpha)/sqrt(1+v))

    current.endpoint <- current.endpoint - c*(y-b)/(beta*b*(1-b))

    v <- v - c^2/(b*(1-b))

    c <- k*abs(current.endpoint - estimate)

    beta <- 1/(c*(dnorm(qnorm(alpha))))

  }

  return(current.endpoint)
}

super_adaptive_regression <- function(estimate, range_l, range_u, start, alpha, steps, sampling_func)
{
  a = 0.05827
  b = 0.05697

  starting.points <- 10
  #middle <- rnorm(1, mean=start)
  middle <- start
  points <- seq(range_l, range_u, length.out=starting.points)

  bootstrap.samples <-  sapply(points, FUN=sampling_func)

  fit <- rq(bootstrap.samples ~ points, tau=alpha)

  intercept = fit$coefficients[1]
  slope = fit$coefficients[2]

  current_endpoint <- (estimate-intercept)/slope

  for(i in 1:(steps-starting.points))
  {
    x3 <- sum(points^3)
    x2 <- sum(points^2)
    x <- sum(points)

    k <- a^2
    l <- 2*a*b-current_endpoint*(a^2)
    m <- b^2-2*a*b*current_endpoint
    n <- (a^2)*x3 + (2*a*b-current_endpoint*(a^2))*x2 + (b^2-2*a*b*current_endpoint)*x - current_endpoint*(length(points)+1)*(b^2)
    roots <- polyroot(c(n,m,l,k))
    possible.points <- Re(roots)[abs(Im(roots)) < 1e-6]

    new.point <- possible.points[sample(1:length(possible.points), 1)]
    new.point <- min(range_u, new.point)
    new.point <- max(range_l, new.point)

    #cat("new point: ", new.point, '\n')

    new.bootstrap.sample <- sampling_func(new.point)

    points <- c(points, new.point)
    bootstrap.samples <- c(bootstrap.samples, new.bootstrap.sample)


    fit <- rq(bootstrap.samples ~ points, tau=alpha)

    intercept = fit$coefficients[1]
    slope = fit$coefficients[2]

    #cat("current endpoint is: ", current_endpoint, '\n')
  }

  plot(bootstrap.samples ~ points)
  return(current_endpoint)
}
